--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        xmake.lua
--

-- define toolchain
toolchain("fpc")

    -- set homepage
    set_homepage("https://www.freepascal.org/")
    set_description("Free Pascal Programming Language Compiler")

    -- set toolset
    set_toolset("pc",   "$(env PC)", "fpc")
    set_toolset("pcld", "$(env PC)", "fpc")
    set_toolset("pcsh", "$(env PC)", "fpc")

    -- on load
    on_load(function (toolchain)
        if toolchain:is_plat("linux") then
            toolchain:set("pcldflags", "-k-lc")
            toolchain:set("pcshflags", "-k-lc")
        else
            toolchain:set("pcldflags", "")
            toolchain:set("pcshflags", "")
        end
    end)
